import re

def sanitize_surrogates(text: str) -> str:
    """
    Removes unpaired Unicode surrogate characters from a string.
    
    Unpaired surrogates (high surrogates 0xD800-0xDBFF without matching low surrogates 0xDC00-0xDFFF,
    or vice versa) cause JSON serialization errors in many API providers.
    
    Valid emoji and other characters outside the Basic Multilingual Plane use properly paired
    surrogates and will NOT be affected by this function.
    
    Args:
        text (str): The text to sanitize
        
    Returns:
        str: The sanitized text with unpaired surrogates removed
        
    Example:
        >>> sanitize_surrogates("Hello \ud83d\ude48 World") 
        'Hello 🙈 World'
        >>> sanitize_surrogates("Text \ud83d here") 
        'Text  here'
    """
    if not text:
        return text
        
    # Replace unpaired high surrogates (0xD800-0xDBFF not followed by low surrogate)
    # Replace unpaired low surrogates (0xDC00-0xDFFF not preceded by high surrogate)
    return re.sub(
        r'[\ud800-\udbff](?![\udc00-\udfff])|(?<![\ud800-\udbff])[\udc00-\udfff]',
        '',
        text
    )



from .sanitize_unicode import sanitize_surrogates
from .partial_json import parse_streaming_json
from .validation import validate_tool_call, validate_tool_arguments

__all__ = [
    "sanitize_surrogates",
    "parse_streaming_json",
    "validate_tool_call",
    "validate_tool_arguments"
]



"""
API provider registry.
"""
from __future__ import annotations
from typing import Any, Optional, Protocol

from .types import ModelDef, Context, StreamEvent, ReasoningEffort, Api
from .providers.openai_completions import OpenAIProvider
from .providers.openai_responses import OpenAIResponsesProvider
from .providers.openai_codex import OpenAICodexProvider
from .providers.gemini_cli import GeminiCliProvider
from .providers.google import GoogleProvider
from .providers.amazon_bedrock import BedrockProvider
from .providers.google_vertex import GoogleVertexProvider

# Define protocols for stream functions
class StreamFunction(Protocol):
    def __call__(
        self,
        model: ModelDef,
        context: Context,
        reasoning: ReasoningEffort = ReasoningEffort.NONE,
    ) -> Any: # Returns AsyncIterator[StreamEvent]
        ...

class ApiProvider:
    def __init__(self, api: Api, stream: StreamFunction, stream_simple: Optional[StreamFunction] = None):
        self.api = api
        self.stream = stream
        self.stream_simple = stream_simple or stream

_registry: dict[str, ApiProvider] = {}

_API_ALIASES: dict[str, str] = {
    # Generated/static model catalog aliases
    "anthropic-messages": "anthropic",
    "openai-completions": "openai",
    "openai-chat-completions": "openai",
    "google-generative-ai": "google",
}

def register_api_provider(provider: ApiProvider) -> None:
    """Register an API provider."""
    if not getattr(provider, "api", None):
        raise ValueError("ApiProvider.api must be a non-empty string")
    if not callable(getattr(provider, "stream", None)):
        raise TypeError("ApiProvider.stream must be callable")
    if getattr(provider, "stream_simple", None) is not None and not callable(provider.stream_simple):
        raise TypeError("ApiProvider.stream_simple must be callable when provided")

    _registry[provider.api] = provider

def get_api_provider(api: str) -> Optional[ApiProvider]:
    """Get a registered API provider."""
    provider = _registry.get(api)
    if provider:
        return provider

    canonical_api = _API_ALIASES.get(api)
    if canonical_api:
        return _registry.get(canonical_api)

    return None

def clear_api_providers() -> None:
    """Clear all registered providers."""
    _registry.clear()



import pytest

from sentarc_ai.providers.openai_responses import OpenAIResponsesProvider
from sentarc_ai.types import (
    Context,
    Message,
    ModelDef,
    ReasoningEffort,
    Role,
    StopEvent,
    StreamOptions,
    TokenUsage,
)


@pytest.mark.asyncio
async def test_openai_responses_uses_options_api_key_and_reasoning(monkeypatch):
    captured = {"api_key": None, "params": None}

    class DummyResponses:
        async def create(self, **params):
            captured["params"] = params

            async def _stream():
                if False:
                    yield None

            return _stream()

    class DummyClient:
        def __init__(self, api_key, base_url=None, default_headers=None):
            captured["api_key"] = api_key
            self.responses = DummyResponses()

    async def fake_process_responses_stream(_stream_resp, _model):
        yield StopEvent(stop_reason="end_turn", usage=TokenUsage())

    monkeypatch.setattr("sentarc_ai.providers.openai_responses.AsyncOpenAI", DummyClient)
    monkeypatch.setattr(
        "sentarc_ai.providers.openai_responses.process_responses_stream",
        fake_process_responses_stream,
    )
    monkeypatch.setattr(
        "sentarc_ai.providers.openai_responses.get_env_api_key",
        lambda _provider: None,
    )

    provider = OpenAIResponsesProvider()
    model = ModelDef(id="o4-mini", provider="openai", api="openai-responses")
    context = Context(messages=[Message(role=Role.USER, content="hello")])
    options = StreamOptions(
        api_key="test-options-key",
        reasoning_effort=ReasoningEffort.HIGH,
        session_id="sess-1",
    )

    events = [e async for e in provider.stream(model, context, options)]

    assert len(events) == 1
    assert events[0].type == "stop"
    assert captured["api_key"] == "test-options-key"
    assert captured["params"]["prompt_cache_key"] == "sess-1"
    assert captured["params"]["reasoning"]["effort"] == "high"



"""
Shared utilities for OpenAI Responses and Codex implementations.
"""
from __future__ import annotations
import json
import re
from typing import Any, AsyncIterator, Callable, Optional, Set, List, Dict, Union

from ..types import (
    Context, ModelDef, StreamEvent, Role, TokenUsage,
    StartEvent, TextStartEvent, TextDeltaEvent, TextEndEvent,
    ThinkingStartEvent, ThinkingDeltaEvent, ThinkingEndEvent,
    ToolUseStartEvent, ToolUseDeltaEvent, ToolUseEndEvent,
    StopEvent, ErrorEvent, ToolUseContent, ToolResultContent,
    AssistantMessage, ToolResultMessage, Message, ContentBlock,
    TextContent, ThinkingContent, ToolCallContent, ImageContent
)
from ..transform_messages import transform_messages

def short_hash(s: str) -> str:
    """Fast deterministic hash to shorten long strings."""
    h1 = 0xdeadbeef
    h2 = 0x41c6ce57
    for char in s:
        ch = ord(char)
        h1 = (h1 ^ ch) * 2654435761 & 0xffffffff
        h2 = (h2 ^ ch) * 1597334677 & 0xffffffff
    
    h1 = ((h1 ^ (h1 >> 16)) * 2246822507) ^ ((h2 ^ (h2 >> 13)) * 3266489909) & 0xffffffff
    h2 = ((h2 ^ (h2 >> 16)) * 2246822507) ^ ((h1 ^ (h1 >> 13)) * 3266489909) & 0xffffffff
    
    # Simple base36-like conversion for brevity in Python
    def to_base36(num):
        chars = "0123456789abcdefghijklmnopqrstuvwxyz"
        if num == 0: return "0"
        res = ""
        while num > 0:
            res = chars[num % 36] + res
            num //= 36
        return res

    return to_base36(h2) + to_base36(h1)

def convert_responses_messages(
    model: ModelDef,
    context: Context,
    allowed_tool_call_providers: Set[str],
    include_system_prompt: bool = True
) -> List[Dict[str, Any]]:
    """Convert context messages to OpenAI Responses API format."""
    messages = []
    
    def normalize_tool_call_id(id_val: str, _model: ModelDef, _msg: AssistantMessage) -> str:
        if model.provider not in allowed_tool_call_providers:
            return id_val
        if "|" not in id_val:
            return id_val
        
        call_id, item_id = id_val.split("|", 1)
        sanitized_call_id = re.sub(r"[^a-zA-Z0-9_-]", "_", call_id)
        sanitized_item_id = re.sub(r"[^a-zA-Z0-9_-]", "_", item_id)
        
        if not sanitized_item_id.startswith("fc"):
            sanitized_item_id = f"fc_{sanitized_item_id}"
            
        norm_call_id = sanitized_call_id[:64].rstrip("_")
        norm_item_id = sanitized_item_id[:64].rstrip("_")
        
        return f"{norm_call_id}|{norm_item_id}"

    transformed = transform_messages(
        context.messages, 
        model, 
        normalize_tool_call_id=normalize_tool_call_id
    )

    if include_system_prompt and context.system_prompt:
        # OpenAI "developer" role for reasoning models, "system" otherwise
        # In this specific context (Responses API), "developer" seems standard for newer models?
        # The TS code uses model.reasoning check.
        # Assuming model.supports_thinking or similar flag implies reasoning capability.
        # Check types.py for ModelDef properties. 
        # Using "developer" if supports_thinking, else "system" as a heuristic.
        role = "developer" if model.supports_thinking else "system"
        messages.append({
            "role": role,
            "content": context.system_prompt
        })

    msg_index = 0
    for msg in transformed:
        if msg.role == Role.USER:
            if isinstance(msg.content, str):
                messages.append({
                    "role": "user",
                    "content": [{"type": "input_text", "text": msg.content}]
                })
            else:
                content_parts = []
                for item in msg.content:
                    if isinstance(item, TextContent):
                        content_parts.append({
                            "type": "input_text",
                            "text": item.text
                        })
                    elif isinstance(item, ImageContent):
                        if model.supports_vision:
                            content_parts.append({
                                "type": "input_image",
                                "detail": "auto",
                                "image_url": f"data:{item.media_type};base64,{item.data}"
                            })
                
                if content_parts:
                    messages.append({
                        "role": "user",
                        "content": content_parts
                    })
                    
        elif msg.role == Role.ASSISTANT:
            output_parts = []
            assistant_msg: AssistantMessage = msg
            
            # Check for different model provenance
            is_different_model = (
                assistant_msg.model != model.id and 
                assistant_msg.provider == model.provider and
                assistant_msg.api == model.api
            )
            
            # Message content in Python is list[ContentBlock]
            # ContentBlock = TextContent | ThinkingContent | ToolCallContent | ...
            
            if isinstance(msg.content, list):
                for block in msg.content:
                    if isinstance(block, ThinkingContent):
                        if block.signature:
                             try:
                                 reasoning_item = json.loads(block.signature)
                                 output_parts.append(reasoning_item)
                             except:
                                 pass
                    elif isinstance(block, TextContent):
                        # OpenAI Responses API uses 'message' items for text output
                        msg_id = f"msg_{msg_index}"
                        if not msg_id:
                            msg_id = f"msg_{msg_index}"
                        elif len(msg_id) > 64:
                            msg_id = f"msg_{short_hash(msg_id)}"
                        
                        output_parts.append({
                            "type": "message",
                            "role": "assistant",
                            "content": [{"type": "output_text", "text": block.text}],
                            "status": "completed",
                            "id": msg_id
                        })
                    elif isinstance(block, ToolCallContent):
                        call_id_raw = block.id.split("|")
                        call_id = call_id_raw[0]
                        item_id = call_id_raw[1] if len(call_id_raw) > 1 else None
                        
                        # Strip itemId if from different model to avoid validation errors
                        if is_different_model and item_id and item_id.startswith("fc_"):
                            item_id = None
                            
                        output_parts.append({
                            "type": "function_call",
                            "id": item_id,
                            "call_id": call_id,
                            "name": block.name,
                            "arguments": json.dumps(block.arguments)
                        })
            
            if output_parts:
                messages.extend(output_parts)
                
        elif msg.role == Role.TOOL:
             # ToolResultMessage
             msg: ToolResultMessage
             
             text_parts = [c.text for c in msg.content if isinstance(c, TextContent)]
             text_result = "\n".join(text_parts)
             has_images = any(isinstance(c, ImageContent) for c in msg.content)
             
             has_text = len(text_result) > 0
             call_id_parts = msg.tool_call_id.split("|")
             call_id = call_id_parts[0]
             
             messages.append({
                 "type": "function_call_output",
                 "call_id": call_id,
                 "output": text_result if has_text else "(see attached image)"
             })
             
             if has_images and model.supports_vision:
                 content_parts = [
                     {"type": "input_text", "text": "Attached image(s) from tool result:"}
                 ]
                 for block in msg.content:
                     if isinstance(block, ImageContent):
                         content_parts.append({
                             "type": "input_image",
                             "detail": "auto",
                             "image_url": f"data:{block.media_type};base64,{block.data}"
                         })
                 messages.append({
                     "role": "user",
                     "content": content_parts
                 })
                 
        msg_index += 1
        
    return messages

def convert_responses_tools(tools: List[Any], options: Optional[Dict] = None) -> List[Dict]:
    """Convert tools to OpenAI format."""
    strict = options.get("strict", False) if options else False
    return [{
        "type": "function",
        "name": t.name,
        "description": t.description,
        "parameters": t.parameters,
        "strict": strict
    } for t in tools]


async def process_responses_stream(
    openai_stream: AsyncIterator[Any],
    model: ModelDef
) -> AsyncIterator[StreamEvent]:
    """Process OpenAI Responses API stream."""
    
    current_item_type = None # "reasoning", "message", "function_call"
    current_block_type = None # "thinking", "text", "toolCall"
    current_text = ""
    current_thinking = ""
    current_tool_call: Optional[Dict] = None
    
    # Store summaries for thinking blocks
    current_thinking_summary = [] 
    
    yield StartEvent(model=model.id)

    def _get_value(obj: Any, key: str, default: Any = None) -> Any:
        if isinstance(obj, dict):
            return obj.get(key, default)
        return getattr(obj, key, default)
    
    try:
        async for event in openai_stream:
            # Event structure in python SDK might differ slightly from TS raw SSE
            # Assuming openai_stream yields objects with 'type' attribute 
            # Or dictionaries if we use raw response parsing (which we might need for Responses API beta)
            
            # The python SDK for Responses API is experimental. 
            # If using `client.responses.create(stream=True)`, it yields events.
            
            # Let's assume event is a dictionary or object with attributes.
            # Using dictionary access for safety if it's raw JSON.
            
            e_type = _get_value(event, "type")
            
            if e_type == "response.output_item.added":
                item = _get_value(event, "item", {})
                i_type = item.get("type")
                
                if i_type == "reasoning":
                    current_item_type = "reasoning"
                    current_block_type = "thinking"
                    current_thinking = ""
                    current_thinking_summary = []
                    yield ThinkingStartEvent()
                    
                elif i_type == "message":
                    current_item_type = "message"
                    current_block_type = "text"
                    current_text = ""
                    yield TextStartEvent()
                    
                elif i_type == "function_call":
                    current_item_type = "function_call"
                    current_block_type = "toolCall"
                    current_tool_call = {
                        "id": f"{item.get('call_id')}|{item.get('id')}",
                        "name": item.get('name'),
                        "args": "",
                        "partial_json": item.get('arguments', "")
                    }
                    yield ToolUseStartEvent(id=current_tool_call["id"], name=current_tool_call["name"])
            
            elif e_type == "response.reasoning_summary_part.added":
                # Only tracking internal state, no yield
                if current_item_type == "reasoning":
                    part = _get_value(event, "part")
                    current_thinking_summary.append(part)

            elif e_type == "response.reasoning_summary_text.delta":
                 if current_item_type == "reasoning" and current_block_type == "thinking":
                     delta = _get_value(event, "delta", "")
                     if delta:
                         current_thinking += delta
                         # Also update last summary part text internally? 
                         yield ThinkingDeltaEvent(thinking=delta)
                         
            elif e_type == "response.reasoning_summary_part.done":
                 if current_item_type == "reasoning":
                     # Add double newline separator
                     yield ThinkingDeltaEvent(thinking="\n\n")

            elif e_type == "response.content_part.added":
                 # message content part added (text or refusal)
                 pass

            elif e_type == "response.output_text.delta":
                 if current_item_type == "message" and current_block_type == "text":
                     delta = _get_value(event, "delta", "")
                     if delta:
                         current_text += delta
                         yield TextDeltaEvent(text=delta)

            elif e_type == "response.refusal.delta":
                 if current_item_type == "message" and current_block_type == "text":
                     delta = _get_value(event, "delta", "")
                     if delta:
                         current_text += delta
                         yield TextDeltaEvent(text=delta)

            elif e_type == "response.function_call_arguments.delta":
                 if current_item_type == "function_call" and current_block_type == "toolCall":
                     delta = getattr(event, "delta", "") or event.get("delta", "")
                     if delta:
                         current_tool_call["partial_json"] += delta
                         yield ToolUseDeltaEvent(id=current_tool_call["id"], partial_input=delta)

            elif e_type == "response.function_call_arguments.done":
                 if current_item_type == "function_call":
                     args = _get_value(event, "arguments", "")
                     current_tool_call["partial_json"] = args

            elif e_type == "response.output_item.done":
                 item = _get_value(event, "item", {})
                 i_type = item.get("type")
                 
                 if i_type == "reasoning" and current_block_type == "thinking":
                     # Finalize thinking
                     yield ThinkingEndEvent(thinking=current_thinking)
                     current_block_type = None
                     
                 elif i_type == "message" and current_block_type == "text":
                     yield TextEndEvent(text=current_text)
                     current_block_type = None
                     
                 elif i_type == "function_call":
                     # Parse args
                     args_str = current_tool_call.get("partial_json", "{}") if current_tool_call else "{}"
                     try:
                         args = json.loads(args_str)
                     except:
                         args = {}

                     tool_id = current_tool_call.get("id", "call_unknown") if current_tool_call else "call_unknown"
                     tool_name = current_tool_call.get("name", "unknown") if current_tool_call else "unknown"
                     
                     yield ToolUseEndEvent(tool_use=ToolUseContent(
                         id=tool_id,
                         name=tool_name,
                         arguments=args
                     ))
                     current_block_type = None

            elif e_type == "response.completed":
                 # Handle usage and finish
                 resp = _get_value(event, "response", {})
                 usage_data = resp.get("usage", {})
                 if usage_data:
                     usage = TokenUsage(
                         input_tokens=usage_data.get("input_tokens", 0),
                         output_tokens=usage_data.get("output_tokens", 0),
                         cache_read_tokens=usage_data.get("input_tokens_details", {}).get("cached_tokens", 0),
                     )
                     # Check status for stop reason
                     status = resp.get("status")
                     stop_reason = "end_turn"
                     if status == "incomplete": stop_reason = "max_tokens"
                     elif status == "failed": stop_reason = "error" # Should have been caught earlier?
                     
                     yield StopEvent(stop_reason=stop_reason, usage=usage)
                     return

            elif e_type == "error":
                 # Stream error
                 code = _get_value(event, "code", "")
                 msg = _get_value(event, "message", "")
                 raise RuntimeError(f"Error {code}: {msg}")

    except Exception as e:
        raise RuntimeError(str(e))



import json
from typing import Any, Dict, List
import jsonschema
from jsonschema import validate, ValidationError

from ..types import Tool, ToolUseContent, ToolCallContent

ToolCallType = ToolUseContent | ToolCallContent

def validate_tool_call(tools: List[Tool], tool_call: ToolCallType) -> Dict[str, Any]:
    """
    Finds a tool by name and validates the tool call arguments against its JSON schema.
    
    Args:
        tools: List of tool definitions.
        tool_call: The tool call from the LLM.
        
    Returns:
        The validated arguments dictionary.
        
    Raises:
        ValueError: If the tool is not found.
        jsonschema.ValidationError: If validation fails, with a formatted message.
    """
    tool = next((t for t in tools if t.name == tool_call.name), None)
    if not tool:
        raise ValueError(f'Tool "{tool_call.name}" not found')
        
    return validate_tool_arguments(tool, tool_call)

def validate_tool_arguments(tool: Tool, tool_call: ToolCallType) -> Dict[str, Any]:
    """
    Validates tool call arguments against the tool's JSON schema.
    
    Args:
        tool: The tool definition with JSON schema.
        tool_call: The tool call from the LLM.
        
    Returns:
        The validated arguments dictionary.
        
    Raises:
        jsonschema.ValidationError: With a formatted message if validation fails.
    """
    try:
        # Load string arguments if it's a JSON string
        args = tool_call.arguments
        if isinstance(args, str):
            try:
                args = json.loads(args)
            except json.JSONDecodeError as e:
                raise ValidationError(
                    f'Failed to parse tool arguments as JSON for "{tool.name}": '
                    f"{e.msg} at line {e.lineno} column {e.colno}. "
                    f"Received: {args[:200]}..."
                )
                
        # jsonschema validates in place and returns None if successful
        validate(instance=args, schema=tool.parameters)
        return args
    except ValidationError as e:
        # Format validation errors nicely to match TS behavior
        path = ".".join([str(p) for p in e.path]) if e.path else "root"
        error_list = f"  - {path}: {e.message}"
        
        error_message = (
            f'Validation failed for tool "{tool_call.name}":\n'
            f'{error_list}\n\n'
            f'Received arguments:\n'
            f'{json.dumps(tool_call.arguments, indent=2)}'
        )
        raise ValidationError(error_message)



"""
Google Gemini provider.
"""
from __future__ import annotations
import json
import asyncio
from typing import AsyncIterator, Optional, Any

try:
    import google.generativeai as genai
    from google.generativeai.types import (
        GenerateContentResponse,
        ContentDict,
        PartDict,
    )
except ImportError:
    # Allow import without package for typing/linting, but stream will fail
    genai = None

from ..types import (
    Context, ModelDef, StreamEvent, Role, Message, StreamOptions, ToolUseContent,
    TokenUsage, ReasoningEffort,
    StartEvent, TextStartEvent, TextDeltaEvent, TextEndEvent,
    ThinkingStartEvent, ThinkingDeltaEvent, ThinkingEndEvent,
    ToolUseStartEvent, ToolUseDeltaEvent, ToolUseEndEvent,
    StopEvent, ErrorEvent,
)
from ..env import get_env_api_key
from ..transform_messages import transform_messages

# ... (imports remain same)

def _convert_messages(messages: list[Message]) -> list[ContentDict]:
    """Convert sentarc-ai messages to Google ContentDict format."""
    contents = []
    for m in messages:
        role = "user" if m.role == Role.USER else "model"
        parts = []
        
        if m.role == Role.TOOL:
            # Tool result
            role = "function" 
            parts.append(PartDict(function_response={
                "name": m.tool_call_id, 
                "response": {"content": m.content} 
            }))
        
        elif isinstance(m.content, str):
            parts.append(PartDict(text=m.content))
            
        elif isinstance(m.content, list):
            for block in m.content:
                if isinstance(block, str):
                     parts.append(PartDict(text=block))
                elif hasattr(block, 'type'):
                    if block.type == "text":
                        parts.append(PartDict(text=block.text))
                    elif block.type == "tool_use":
                        parts.append(PartDict(function_call={
                            "name": block.name,
                            "args": block.arguments
                        }))
                    elif block.type == "tool_result":
                        pass

        if parts:
            contents.append(ContentDict(role=role, parts=parts))
            
    return contents
# ... (rest of _convert_messages implementation unchanged, just signature comment updated if needed, but logic uses iteration so fine)



def _convert_tools(tools) -> list[Any]:
    """Convert sentarc-ai tools to Google Tool format."""
    # SDK expects tools=[{'function_declarations': [...]}]
    funcs = []
    for t in tools:
        # Convert JSON schema types to Google types if needed, 
        # or pass raw if SDK supports it (latest versions do allow dicts).
        funcs.append({
            "name": t.name,
            "description": t.description,
            "parameters": t.parameters
        })
    return [{"function_declarations": funcs}]



class GoogleProvider:
    def __init__(self):
        self.api = "google"

    async def stream(
        self,
        model: "ModelDef",
        context: "Context",
        options: Optional[StreamOptions] = None,
    ) -> AsyncIterator[StreamEvent]:
        """Stream from Google Gemini API."""
        if not genai:
            raise RuntimeError("google-generativeai package not found. Install it with pip.")

        api_key = get_env_api_key(model.provider)
        if not api_key:
            raise RuntimeError(f"No API key found for {model.provider}")

        genai.configure(api_key=api_key)
        
        generation_config = {
            "temperature": 0.7,
            "max_output_tokens": model.max_output,
        }
        
        model_tools = _convert_tools(context.tools) if context.tools else None
        
        gemini_model = genai.GenerativeModel(
            model_name=model.id,
            generation_config=generation_config,
            tools=model_tools
        )

        transformed_messages = transform_messages(context.messages, model)
        contents = _convert_messages(transformed_messages)
        
        yield StartEvent(model=model.id)
        
        text_started = False
        
        try:
            response = await gemini_model.generate_content_async(
                contents,
                stream=True,
                generation_config=generation_config
            )
            
            async for chunk in response:
                if not chunk.candidates: continue
                
                parts = chunk.candidates[0].content.parts
                for part in parts:
                    if part.text:
                        if not text_started:
                            yield TextStartEvent()
                            text_started = True
                        yield TextDeltaEvent(text=part.text)
                    
                    if part.function_call:
                        fc = part.function_call
                        import uuid
                        import json
                        tc_id = f"call_{fc.name}_{str(uuid.uuid4())[:8]}"
                        
                        yield ToolUseStartEvent(id=tc_id, name=fc.name)
                        args = dict(fc.args)
                        yield ToolUseDeltaEvent(id=tc_id, partial_input=json.dumps(args))
                        yield ToolUseEndEvent(tool_use=ToolUseContent(id=tc_id, name=fc.name, arguments=args))

            if response.usage_metadata:
               u = response.usage_metadata
               usage = TokenUsage(
                   input_tokens=u.prompt_token_count,
                   output_tokens=u.candidates_token_count,
               )
               yield StopEvent(stop_reason="end_turn", usage=usage)
            else:
               yield StopEvent(stop_reason="end_turn", usage=TokenUsage())

        except Exception as e:
            raise RuntimeError(str(e))



"""
OpenAI Codex Responses API provider.
"""
from __future__ import annotations

import json
from typing import AsyncIterator, Optional, Any, Dict

import httpx

from ..types import Context, ModelDef, StreamEvent
from ..env import get_env_api_key
from .openai_utils import (
    convert_responses_messages,
    convert_responses_tools,
    process_responses_stream,
)

CODEX_TOOL_CALL_PROVIDERS = {"openai", "openai-codex", "opencode"}
DEFAULT_CODEX_BASE_URL = "https://chatgpt.com/backend-api"


class OpenAICodexProvider:
    def __init__(self):
        self.api = "openai-codex-responses"

    async def stream(
        self,
        model: "ModelDef",
        context: "Context",
        options: Optional[Any] = None,
    ) -> AsyncIterator[StreamEvent]:
        """Stream from OpenAI Codex Responses API."""

        resolved_api_key = getattr(options, "api_key", None) if options else None
        api_key = resolved_api_key or get_env_api_key(model.provider)
        if not api_key:
            raise RuntimeError(f"No API key found for {model.provider}")

        messages = convert_responses_messages(
            model,
            context,
            CODEX_TOOL_CALL_PROVIDERS,
            include_system_prompt=False,
        )

        body: Dict[str, Any] = {
            "model": model.id,
            "store": False,
            "stream": True,
            "instructions": context.system_prompt,
            "input": messages,
            "text": {"verbosity": getattr(options, "text_verbosity", "medium") if options else "medium"},
            "include": ["reasoning.encrypted_content"],
            "tool_choice": "auto",
            "parallel_tool_calls": True,
        }

        if options and getattr(options, "session_id", None):
            body["prompt_cache_key"] = options.session_id

        if context.tools:
            body["tools"] = convert_responses_tools(context.tools, options={"strict": None})

        base_url = model.base_url or DEFAULT_CODEX_BASE_URL
        url = base_url.rstrip("/")
        if url.endswith("/codex/responses"):
            pass
        elif url.endswith("/codex"):
            url += "/responses"
        else:
            url += "/codex/responses"

        account_id = api_key.split("::", 1)[0] if "::" in api_key else None

        headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
            "Accept": "text/event-stream",
        }
        if account_id:
            headers["X-OpenAI-Account-ID"] = account_id
        if model.extra_headers:
            headers.update(model.extra_headers)

        try:
            event_generator = self._fetch_events(url, body, headers)
            async for event in process_responses_stream(event_generator, model):
                yield event
        except Exception as e:
            raise RuntimeError(str(e))

    async def _fetch_events(self, url: str, body: Dict[str, Any], headers: Dict[str, str]) -> AsyncIterator[Dict[str, Any]]:
        async with httpx.AsyncClient(timeout=120.0) as client:
            async with client.stream("POST", url, json=body, headers=headers) as response:
                if response.status_code != 200:
                    err_text = await response.aread()
                    yield {"type": "error", "code": str(response.status_code), "message": err_text.decode("utf-8")}
                    return

                async for line in response.aiter_lines():
                    if not line.startswith("data:"):
                        continue
                    data_str = line[5:].strip()
                    if not data_str or data_str == "[DONE]":
                        continue
                    try:
                        yield json.loads(data_str)
                    except Exception:
                        continue




import os
import json
import time
import boto3
from typing import AsyncIterator, Optional, Any, Literal
from ..types import (
    ModelDef, Context, AssistantMessage,
    StreamEvent, TextContent, ToolCallContent as ToolCall, ToolResultContent,
    StreamOptions, ReasoningEffort as ThinkingLevel, Message
)
from ..utils.partial_json import parse_streaming_json
from ..transform_messages import transform_messages

def map_thinking_level_to_budget(level: ThinkingLevel, model_id: str) -> int:
    # Default budgets
    defaults = {
        ThinkingLevel.MINIMAL: 1024,
        ThinkingLevel.LOW: 2048,
        ThinkingLevel.MEDIUM: 8192,
        ThinkingLevel.HIGH: 16384,
        ThinkingLevel.XHIGH: 16384, # Clamp xhigh to high for now
    }
    return defaults.get(level, 16384)

class BedrockProvider:
    def __init__(self):
        self.api = "bedrock-converse-stream"

    async def stream(
        self,
        model: ModelDef,
        context: Context,
        options: Optional[StreamOptions] = None,
    ) -> AsyncIterator[StreamEvent]:
        options = options or StreamOptions()
        
        # Initialize boto3 client
        region = os.getenv("AWS_REGION", "us-east-1")
        if "AWS_BEDROCK_REGION" in os.environ:
             region = os.environ["AWS_BEDROCK_REGION"]
             
        client = boto3.client("bedrock-runtime", region_name=region)

        client = boto3.client("bedrock-runtime", region_name=region)

        # Transform & Convert
        transformed = transform_messages(context.messages, model)
        messages = self._convert_messages(transformed)
        
        # Prepare system prompt
        system = []
        if context.system_prompt:
            system.append({"text": context.system_prompt})

        # Inference config
        inference_config = {
            "maxTokens": options.max_tokens or model.max_output or 4096,
            "temperature": options.temperature if options.temperature is not None else 0.7,
        }

        # Tool config
        tool_config = self._convert_tool_config(context.tools)

        # Additional model fields (Thinking)
        additional_model_fields = {}
        if options.thinking_enabled and model.supports_thinking:
             # Check if model supports reasoning/thinking
             # For Claude 3.7 Sonnet / Claude 3.5 Haiku?
             # Implementation matches typescript logic
             if "anthropic.claude" in model.id:
                 budget = options.thinking_budget 
                 if not budget and options.reasoning_effort:
                     budget = map_thinking_level_to_budget(options.reasoning_effort, model.id)
                 
                 if budget:
                    additional_model_fields["thinking"] = {
                        "type": "enabled",
                        "budget_tokens": budget
                    }
                 elif "opus-4-6" in model.id: # Adaptive example
                     pass # handled differently in TS, simplistic here

        args = {
            "modelId": model.id,
            "messages": messages,
            "inferenceConfig": inference_config,
        }
        if system:
            args["system"] = system
        if tool_config:
            args["toolConfig"] = tool_config
        if additional_model_fields:
            args["additionalModelRequestFields"] = additional_model_fields

        # Invoke stream
        response = client.converse_stream(**args)
        stream = response.get("stream")

        output_message = AssistantMessage(
            role="assistant",
            content=[],
            provider=model.provider,
            model=model.id,
        )

        current_block_index = 0
        current_tool_id = ""
        current_tool_name = ""
        current_tool_json = ""
        
        if stream:
            for event in stream:
                if "messageStart" in event:
                    yield StreamEvent(type="start", partial=output_message)
                
                elif "contentBlockStart" in event:
                    start = event["contentBlockStart"]
                    index = start["contentBlockIndex"]
                    if "toolUse" in start["start"]:
                        tool_use = start["start"]["toolUse"]
                        current_tool_id = tool_use["toolUseId"]
                        current_tool_name = tool_use["name"]
                        current_tool_json = ""
                        
                        tool_call = ToolCall(
                            id=current_tool_id,
                            name=current_tool_name,
                            arguments={}
                        )
                        output_message.content.append(tool_call)
                        current_block_index = len(output_message.content) - 1
                        yield StreamEvent(type="toolcall_start", content_index=current_block_index, partial=output_message)
                    elif "text" in start.get("start", {}):
                        # Text block start usually implied or handled in delta
                        pass

                elif "contentBlockDelta" in event:
                    delta = event["contentBlockDelta"]
                    delta_content = delta["delta"]
                    
                    if "text" in delta_content:
                        text = delta_content["text"]
                        
                        # Ensure we have a text block
                        if not output_message.content or not isinstance(output_message.content[-1], TextContent):
                             output_message.content.append(TextContent(text=""))
                             current_block_index = len(output_message.content) - 1
                             yield StreamEvent(type="text_start", content_index=current_block_index, partial=output_message)
                        
                        # Append text
                        block = output_message.content[-1]
                        if isinstance(block, TextContent):
                            block.text += text
                            yield StreamEvent(type="text_delta", content_index=current_block_index, delta=text, partial=output_message)

                    elif "toolUse" in delta_content:
                         input_delta = delta_content["toolUse"]["input"]
                         current_tool_json += input_delta
                         
                         # Update arguments in partial
                         block = output_message.content[current_block_index]
                         if isinstance(block, ToolCall):
                             block.arguments = parse_streaming_json(current_tool_json)
                             yield StreamEvent(type="toolcall_delta", content_index=current_block_index, delta=input_delta, partial=output_message)
                    
                    elif "reasoningContent" in delta_content:
                        # Handle thinking/reasoning
                        # TODO: Implement ThinkingContent support in types.py if not exists
                         pass 

                elif "contentBlockStop" in event:
                    stop = event["contentBlockStop"]
                    index = stop["contentBlockIndex"]
                    block = output_message.content[len(output_message.content) - 1] # Simplified assumption
                    
                    if isinstance(block, ToolCall):
                         block.arguments = parse_streaming_json(current_tool_json)
                         yield StreamEvent(type="toolcall_end", content_index=current_block_index, tool_call=block, partial=output_message)
                    elif isinstance(block, TextContent):
                         yield StreamEvent(type="text_end", content_index=current_block_index, content=block.text, partial=output_message)

                elif "messageStop" in event:
                    stop_reason = event["messageStop"]["stopReason"]
                    reason_map = {
                        "end_turn": "stop",
                        "tool_use": "toolUse",
                        "max_tokens": "length",
                        "stop_sequence": "stop"
                    }
                    output_message.stop_reason = reason_map.get(stop_reason, "stop")

                elif "metadata" in event:
                    usage = event["metadata"]["usage"]
                    output_message.usage.input_tokens = usage["inputTokens"]
                    output_message.usage.output_tokens = usage["outputTokens"]
                    output_message.usage.total_tokens = usage["totalTokens"]
        
        yield StreamEvent(type="done", reason=output_message.stop_reason, message=output_message)


    def _convert_messages(self, messages: list[Message]) -> list[dict]:
        bedrock_messages = []
        for m in messages:
            role = "user" if m.role == "user" else "assistant"
            content = []
            
            if isinstance(m.content, str):
                content.append({"text": m.content})
            elif isinstance(m.content, list):
                for block in m.content:
                    if isinstance(block, TextContent):
                         content.append({"text": block.text})
                    elif isinstance(block, ToolCall):
                         content.append({
                             "toolUse": {
                                 "toolUseId": block.id,
                                 "name": block.name,
                                 "input": block.arguments
                             }
                         })
                    elif isinstance(block, ToolResultContent):
                         # Bedrock expects tool results in a separate structure? 
                         # Actually converse API handles 'user' role with 'toolResult' content blocks
                         pass

            # Handle Tool Results (which are "user" role in Bedrock Converse)
            if m.role == "toolResult":
                 role = "user"
                 content = [{
                     "toolResult": {
                         "toolUseId": m.tool_call_id,
                         "content": [{"text": c.text} for c in m.content if isinstance(c, TextContent)],
                         "status": "error" if m.is_error else "success"
                     }
                 }]
            
            bedrock_messages.append({"role": role, "content": content})
        return bedrock_messages

    def _convert_tool_config(self, tools: Optional[list]) -> Optional[dict]:
        if not tools:
            return None
        
        bedrock_tools = []
        for tool in tools:
            bedrock_tools.append({
                "toolSpec": {
                    "name": tool.name,
                    "description": tool.description,
                    "inputSchema": {"json": tool.parameters}
                }
            })
        return {"tools": bedrock_tools} 



import pytest

from sentarc_ai.providers.openai_completions import OpenAIProvider
from sentarc_ai.types import Context, Message, ModelDef, Role, StreamOptions, TokenUsage, StopEvent


@pytest.mark.asyncio
async def test_openai_completions_uses_options_api_key_and_ignores_missing_tool_choice(monkeypatch):
    captured = {"api_key": None}

    class DummyChunk:
        def __init__(self):
            class Choice:
                def __init__(self):
                    class Delta:
                        content = None
                        tool_calls = None
                    self.delta = Delta()
            self.choices = [Choice()]
            self.usage = type("Usage", (), {"prompt_tokens": 1, "completion_tokens": 2, "total_tokens": 3})

    class DummyResponseStream:
        def __aiter__(self):
            return self

        async def __anext__(self):
            if getattr(self, "_done", False):
                raise StopAsyncIteration
            self._done = True
            return DummyChunk()

    class DummyCompletions:
        async def create(self, **kwargs):
            return DummyResponseStream()

    class DummyChat:
        def __init__(self):
            self.completions = DummyCompletions()

    class DummyClient:
        def __init__(self, api_key, base_url=None, default_headers=None):
            captured["api_key"] = api_key
            self.chat = DummyChat()

    monkeypatch.setattr("sentarc_ai.providers.openai_completions.AsyncOpenAI", DummyClient)
    monkeypatch.setattr("sentarc_ai.providers.openai_completions.get_env_api_key", lambda _provider: None)

    provider = OpenAIProvider()
    model = ModelDef(id="gpt-4o", provider="openai", api="openai")
    context = Context(messages=[Message(role=Role.USER, content="hello")])
    options = StreamOptions(api_key="opt-key")

    events = [e async for e in provider.stream(model, context, options)]

    assert captured["api_key"] == "opt-key"
    assert any(getattr(e, "type", None) == "stop" for e in events)



"""
Message transformation — packages/ai/src/transform-messages.ts from sentarc-mono.
"""
from __future__ import annotations
import json
import time
from typing import Optional, Callable, Any
from dataclasses import replace
from .types import (
    Message, Role, TextContent, ImageContent, ToolCallContent, 
    AssistantMessage, ToolResultMessage, ModelDef, ContentBlock,
    TokenUsage, ThinkingContent, ToolUseContent
)
from .utils.sanitize_unicode import sanitize_surrogates

def transform_messages(
    messages: list[Message],
    model: ModelDef,
    normalize_tool_call_id: Optional[Callable[[str, ModelDef, AssistantMessage], str]] = None,
) -> list[Message]:
    """
    Transform messages for the target model.
    Handles:
    - Thinking blocks (strip/keep based on model support)
    - Tool call ID normalization (cross-provider)
    - Synthetic tool results for orphaned tool calls
    - filtering out error/aborted messages
    """
    
    # Map original tool call IDs to normalized IDs
    tool_call_id_map: dict[str, str] = {}
    
    # -------------------------------------------------------
    # Pass 1: Transform content (thinking, tool IDs)
    # -------------------------------------------------------
    transformed: list[Message] = []
    
    for msg in messages:
        # Sanitize system prompts or generic string content
        if msg.role in (Role.USER, Role.SYSTEM) and isinstance(msg.content, str):
            transformed.append(replace(msg, content=sanitize_surrogates(msg.content)))
            continue
            
        # User messages with blocks
        if msg.role == Role.USER:
            if isinstance(msg.content, list):
                new_blocks = []
                for block in msg.content:
                    if isinstance(block, TextContent):
                        new_blocks.append(replace(block, text=sanitize_surrogates(block.text)))
                    else:
                        new_blocks.append(block)
                transformed.append(replace(msg, content=new_blocks))
            else:
                 transformed.append(msg)
            continue
            
        # Tool results - check if ID needs remapping
        if msg.role == Role.TOOL:
            # Cast to ToolResultMessage if possible, else treat as generic Message
            tool_id = msg.tool_call_id or "" 
            
            normalized_id = tool_call_id_map.get(tool_id)
            if normalized_id and normalized_id != tool_id:
                # Create a copy with new ID using ToolResultMessage if applicable or generic
                new_msg = ToolResultMessage(
                    role=Role.TOOL,
                    content=msg.content,
                    tool_call_id=normalized_id,
                    is_error=getattr(msg, "is_error", False),
                    timestamp=getattr(msg, "timestamp", None)
                )
                transformed.append(new_msg)
            else:
                transformed.append(msg)
            continue
            
        # Assistant messages
        if msg.role == Role.ASSISTANT:
            # Check stop reason - skip if error/aborted
            stop_reason = getattr(msg, "stop_reason", None)
            if stop_reason in ["error", "aborted"]:
                continue
                
            # Determine if this message is from the same model
            msg_provider = getattr(msg, "provider", None)
            msg_model = getattr(msg, "model", None)
            
            # If provider/model not set (generic Message), assume same model or unknown
            # If unknown, we usually treat as same unless restricted?
            # TypeScript logic: isSameModel = (msg.provider === model.provider && ...)
            # If msg is generic, these are None. 
            # We'll treat None as "different" or "generic"? 
            # Let's assume generic messages (from user input history) might need processing.
            
            is_same_model = (
                msg_provider == model.provider and
                msg_model == model.id
            )
            
            # Transform content blocks
            new_content: list[ContentBlock] = []
            
            # Normalize content to list of blocks
            blocks = []
            if isinstance(msg.content, str):
                blocks = [TextContent(text=msg.content)]
            elif isinstance(msg.content, list):
                blocks = msg.content
            
            for block in blocks:
                if isinstance(block, ThinkingContent):
                    # Sanitize thinking block strings
                    sanitized_thinking = sanitize_surrogates(block.thinking)
                    # Same model: keep thinking (with signature)
                    if is_same_model and block.signature:
                        new_content.append(replace(block, thinking=sanitized_thinking))
                        continue
                        
                    if not sanitized_thinking or not sanitized_thinking.strip():
                         if is_same_model and block.signature: # Encrypted reasoning
                             new_content.append(replace(block, thinking=sanitized_thinking))
                         continue
                        
                    if is_same_model:
                        new_content.append(replace(block, thinking=sanitized_thinking))
                    else:
                        # Convert to text if different model
                        new_content.append(TextContent(text=sanitized_thinking))
                        
                elif isinstance(block, TextContent):
                    new_content.append(replace(block, text=sanitize_surrogates(block.text)))
                    
                elif isinstance(block, (ToolCallContent, ToolUseContent)):
                    # Normalize tool call
                    tool_id = block.id
                    args = block.arguments
                    name = block.name
                    
                    final_id = tool_id
                    
                    if not is_same_model and normalize_tool_call_id:
                        # Create a temp AssistantMessage wrapper for context if needed
                        # We pass the original msg (or cast it)
                        normalized = normalize_tool_call_id(tool_id, model, msg) # type: ignore
                        if normalized != tool_id:
                            tool_call_id_map[tool_id] = normalized
                            final_id = normalized
                            
                    # Reconstruct block
                    if isinstance(block, ToolCallContent):
                        new_content.append(ToolCallContent(id=final_id, name=name, arguments=args))
                    else:
                         new_content.append(ToolUseContent(id=final_id, name=name, arguments=args))
                         
                else:
                    new_content.append(block)
            
            # Update tool_calls list on the message for convenience
            new_tool_calls = [b for b in new_content if isinstance(b, (ToolCallContent, ToolUseContent))]
            
            # Create new assistant message
            new_msg = AssistantMessage(
                role=Role.ASSISTANT,
                content=new_content,
                tool_calls=new_tool_calls, # type: ignore
                stop_reason=stop_reason,
                usage=getattr(msg, "usage", None) or TokenUsage(),
                provider=msg_provider,
                model=msg_model
            )
            transformed.append(new_msg)
            continue
            
        # Fallback for SYSTEM or other containing str
        if isinstance(msg.content, str):
            transformed.append(replace(msg, content=sanitize_surrogates(msg.content)))
        else:
            transformed.append(msg)
        
        
    # -------------------------------------------------------
    # Pass 2: Insert synthetic tool results for orphans
    # -------------------------------------------------------
    final_messages: list[Message] = []
    pending_tool_calls: list[ToolCallContent | ToolUseContent] = []
    existing_tool_result_ids: set[str] = set()
    
    for msg in transformed:
        if msg.role == Role.ASSISTANT:
            # If we have pending orphans, close them out
            if pending_tool_calls:
                for tc in pending_tool_calls:
                    if tc.id not in existing_tool_result_ids:
                        final_messages.append(ToolResultMessage(
                            role=Role.TOOL,
                            tool_call_id=tc.id,
                            content="No result provided",
                            is_error=True,
                            timestamp=time.time()
                        ))
                pending_tool_calls = []
                existing_tool_result_ids = set()
            
            # Track new tool calls
            # msg.tool_calls is populated on AssistantMessage
            if msg.tool_calls:
                 pending_tool_calls = msg.tool_calls
                 existing_tool_result_ids = set()
            
            final_messages.append(msg)
            
        elif msg.role == Role.TOOL:
            if msg.tool_call_id:
                existing_tool_result_ids.add(msg.tool_call_id)
            final_messages.append(msg)
            
        elif msg.role == Role.USER:
            # User interrupts tool flow
            if pending_tool_calls:
                for tc in pending_tool_calls:
                     if tc.id not in existing_tool_result_ids:
                        final_messages.append(ToolResultMessage(
                            role=Role.TOOL,
                            tool_call_id=tc.id,
                            content="No result provided",
                            is_error=True,
                            timestamp=time.time()
                        ))
                pending_tool_calls = []
                existing_tool_result_ids = set()
            final_messages.append(msg)
            
        else:
            final_messages.append(msg)
            
    # Final flush
    if pending_tool_calls:
        for tc in pending_tool_calls:
             if tc.id not in existing_tool_result_ids:
                final_messages.append(ToolResultMessage(
                    role=Role.TOOL,
                    tool_call_id=tc.id,
                    content="No result provided",
                    is_error=True,
                    timestamp=time.time()
                ))
                
    return final_messages


# -------------------------------------------------------
# Provider-Specific Formatters (Adapters)
# -------------------------------------------------------

def to_anthropic_messages(messages: list[Message]) -> list[dict]:
    result = []
    for m in messages:
        if m.role == Role.SYSTEM: continue
        
        if m.role == Role.TOOL:
            content = _anthropic_content(m.content)
            # Ensure it's a list for tool_result
            final_content = []
            if isinstance(m.content, list):
                 for b in m.content:
                     if isinstance(b, ToolResultContent):
                         final_content.append({
                             "type": "tool_result",
                             "tool_use_id": b.tool_call_id,
                             "content": b.content, 
                             "is_error": getattr(m, "is_error", False)
                         })
            else:
                 final_content.append({
                     "type": "tool_result",
                     "tool_use_id": m.tool_call_id or "unknown",
                     "content": str(m.content),
                     "is_error": getattr(m, "is_error", False)
                 })
                 
            result.append({"role": "user", "content": final_content})

        elif m.role == Role.ASSISTANT:
            content = []
            if isinstance(m.content, str) and m.content:
                content.append({"type": "text", "text": m.content})
            elif isinstance(m.content, list):
                for b in m.content:
                    if isinstance(b, TextContent):
                        content.append({"type": "text", "text": b.text})
                    elif isinstance(b, ThinkingContent):
                        content.append({
                            "type": "thinking",
                            "thinking": b.thinking,
                            "signature": b.signature
                        })
                    elif isinstance(b, (ToolCallContent, ToolUseContent)):
                        content.append({
                            "type": "tool_use",
                            "id": b.id,
                            "name": b.name,
                            "input": b.arguments
                        })
            
            result.append({"role": "assistant", "content": content})
            
        else:
            # User
            result.append({"role": "user", "content": _anthropic_content(m.content)})
            
    return result

def _anthropic_content(c):
    if isinstance(c, str): return c
    if isinstance(c, list):
        out = []
        for b in c:
            if isinstance(b, TextContent):
                 out.append({"type": "text", "text": b.text})
            elif isinstance(b, ImageContent):
                 out.append({
                     "type": "image",
                     "source": {
                         "type": "base64",
                         "media_type": b.media_type,
                         "data": b.data
                     }
                 })
        return out
    return str(c)


def to_openai_messages(messages: list[Message]) -> list[dict]:
    result = []
    for m in messages:
        if m.role == Role.SYSTEM:
            result.append({"role": "system", "content": _openai_content(m.content)})
            continue
            
        if m.role == Role.TOOL:
            result.append({
                "role": "tool",
                "tool_call_id": m.tool_call_id,
                "content": _openai_content(m.content)
            })
            
        elif m.role == Role.ASSISTANT:
            msg = {"role": "assistant"}
            content = _openai_content(m.content)
            if content:
                msg["content"] = content
                
            if m.tool_calls:
                msg["tool_calls"] = [{
                    "id": tc.id,
                    "type": "function",
                    "function": {
                        "name": tc.name,
                        "arguments": json.dumps(tc.arguments)
                    }
                } for tc in m.tool_calls]
            result.append(msg)
            
        else:
            result.append({"role": "user", "content": _openai_content(m.content)})
            
    return result

def _openai_content(c):
    if isinstance(c, str): return c
    if isinstance(c, list):
        # Flatten simple text blocks?
        # OpenAI supports list of content parts
        parts = []
        for b in c:
            if isinstance(b, TextContent):
                parts.append({"type": "text", "text": b.text})
            elif isinstance(b, ImageContent):
                parts.append({
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:{b.media_type};base64,{b.data}"
                    }
                })
        # If single text, return str? No, keep list for robustness if mixed
        if len(parts) == 1 and parts[0]["type"] == "text":
            return parts[0]["text"]
        return parts if parts else ""
    return str(c)

# Aliases
to_anthropic = to_anthropic_messages
to_openai    = to_openai_messages




"""OpenAI provider — supports both /completions and /responses endpoints."""
from __future__ import annotations
import json
from typing import AsyncIterator, Optional, Any

import openai
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletionChunk

from ..types import (
    Context, ModelDef, StreamEvent, Role,
    TokenUsage, ReasoningEffort,
    StartEvent, TextStartEvent, TextDeltaEvent, TextEndEvent,
    ThinkingStartEvent, ThinkingDeltaEvent, ThinkingEndEvent,
    ToolUseStartEvent, ToolUseDeltaEvent, ToolUseEndEvent,
    StopEvent, ErrorEvent, ToolUseContent, ToolCallContent,
)
from ..env import get_env_api_key
# ... (imports)

from ..transform_messages import to_openai_messages, transform_messages


class OpenAIProvider:
    def __init__(self):
        self.api = "openai"

    async def stream(
        self,
        model: "ModelDef",
        context: "Context",
        options: Optional[Any] = None,
    ) -> AsyncIterator[StreamEvent]:
        """Stream from OpenAI API."""
        
        reasoning = options.reasoning_effort if options else ReasoningEffort.NONE
        
        resolved_api_key = getattr(options, "api_key", None) if options else None
        api_key = resolved_api_key or get_env_api_key(model.provider)
        if not api_key:
            raise RuntimeError(f"No API key found for {model.provider}")

        client = AsyncOpenAI(
            api_key=api_key,
            base_url=model.base_url,
            default_headers=model.extra_headers
        )

        # Convert messages
        transformed = transform_messages(context.messages, model)
        messages = to_openai_messages(transformed)
        
        if context.system_prompt:
            role = "developer" if (model.supports_thinking or model.id.startswith("o")) else "system"
            messages.insert(0, {"role": role, "content": context.system_prompt})

        params = {
            "model": model.id,
            "messages": messages,
            "stream": True,
        }

        if context.tools:
            params["tools"] = [{
                "type": "function",
                "function": {
                    "name": t.name,
                    "description": t.description,
                    "parameters": t.parameters
                }
            } for t in context.tools]
            tool_choice = getattr(context, "tool_choice", None)
            if tool_choice:
                params["tool_choice"] = tool_choice

        if model.supports_thinking and reasoning != ReasoningEffort.NONE:
             params["reasoning_effort"] = reasoning.value

        yield StartEvent(model=model.id)
        
        try:
            stream_resp = await client.chat.completions.create(**params)
            
            text_started = False
            current_tool_call = None
            
            async for chunk in stream_resp:
                if not chunk.choices: continue
                
                delta = chunk.choices[0].delta
                
                # Content / Text
                if delta.content:
                    if not text_started:
                        yield TextStartEvent()
                        text_started = True
                    yield TextDeltaEvent(text=delta.content)
                
                # Reasoning Content
                if hasattr(delta, "reasoning_content") and delta.reasoning_content:
                    yield ThinkingDeltaEvent(thinking=delta.reasoning_content)
                    
                # Tool Calls
                if delta.tool_calls:
                    for tc in delta.tool_calls:
                        if tc.id:
                            current_tool_call = {
                                "index": tc.index, 
                                "id": tc.id, 
                                "name": tc.function.name, 
                                "args": ""
                            }
                            yield ToolUseStartEvent(id=tc.id, name=tc.function.name)
                        
                        if tc.function and tc.function.arguments:
                            if current_tool_call and current_tool_call["index"] == tc.index:
                                current_tool_call["args"] += tc.function.arguments
                                yield ToolUseDeltaEvent(id=current_tool_call["id"], partial_input=tc.function.arguments)

                if chunk.usage:
                    u = chunk.usage
                    usage = TokenUsage(
                        input_tokens=u.prompt_tokens,
                        output_tokens=u.completion_tokens,
                    )
                    yield StopEvent(stop_reason="end_turn", usage=usage)
                    return

            if current_tool_call:
                try:
                    args = json.loads(current_tool_call["args"])
                    yield ToolUseEndEvent(tool_use=ToolCallContent( # Using ToolCallContent from types
                         id=current_tool_call["id"],
                         name=current_tool_call["name"],
                         arguments=args
                    ))
                except json.JSONDecodeError:
                    raise RuntimeError("Failed to parse tool arguments")
            
            if text_started:
                yield TextEndEvent(text="")
                
            yield StopEvent(stop_reason="end_turn", usage=TokenUsage())
            
        except Exception as e:
            raise RuntimeError(str(e))






"""
Unified streaming entry point.

Usage:
    async for event in stream(model, context):
        ...
"""
from __future__ import annotations
from typing import AsyncIterator, Optional, Tuple, Literal

from .types import Context, ModelDef, StreamEvent, StreamOptions, ReasoningEffort, ErrorEvent, AssistantMessage, TextContent, ThinkingContent, ToolUseContent, TokenUsage
from .registry import get_api_provider

async def _with_error_handling(
    generator: AsyncIterator[StreamEvent],
    model: ModelDef
) -> AsyncIterator[StreamEvent]:
    import asyncio
    import json
    from .types import TextContent, ThinkingContent, ToolUseContent, TokenUsage, AssistantMessage, ErrorEvent
    
    blocks = []
    current_text = ""
    current_thinking = ""
    current_tool_id = ""
    current_tool_name = ""
    current_tool_args = ""
    usage = TokenUsage()

    try:
        async for event in generator:
            if event.type == "text_delta":
                current_text += event.text
            elif event.type == "text_end":
                if current_text:
                    blocks.append(TextContent(text=current_text))
                    current_text = ""
            elif event.type == "thinking_delta":
                current_thinking += event.thinking
            elif event.type == "thinking_end":
                if current_thinking:
                    blocks.append(ThinkingContent(thinking=current_thinking))
                    current_thinking = ""
            elif event.type == "tool_use_start":
                current_tool_id = event.id
                current_tool_name = event.name
                current_tool_args = ""
            elif event.type == "tool_use_delta":
                current_tool_args += event.partial_input
            elif event.type == "tool_use_end":
                blocks.append(event.tool_use)
                current_tool_id = ""
                current_tool_name = ""
                current_tool_args = ""
            elif event.type == "stop":
                usage = event.usage

            yield event
            
    except asyncio.CancelledError:
        content = list(blocks)
        if current_text: content.append(TextContent(text=current_text))
        if current_thinking: content.append(ThinkingContent(thinking=current_thinking))
        
        msg = AssistantMessage(
            role="assistant",
            content=content,
            stop_reason="aborted",
            usage=usage,
            provider=model.provider,
            model=model.id,
            api=model.api,
            error_message="Request was cancelled via abort/cancellation signal"
        )
        yield ErrorEvent(error=msg, reason="aborted")
        
    except Exception as e:
        content = list(blocks)
        if current_text: content.append(TextContent(text=current_text))
        if current_thinking: content.append(ThinkingContent(thinking=current_thinking))
        
        msg = AssistantMessage(
            role="assistant",
            content=content,
            stop_reason="error",
            usage=usage,
            provider=model.provider,
            model=model.id,
            api=model.api,
            error_message=str(e)
        )
        yield ErrorEvent(error=msg, reason="error")
    finally:
        if hasattr(generator, "aclose"):
            try:
                await generator.aclose()
            except Exception:
                pass

async def stream(
    model: ModelDef,
    context: Context,
    options: Optional[StreamOptions] = None,
) -> AsyncIterator[StreamEvent]:
    """
    Unified streaming interface across all providers.
    Yields StreamEvent objects normalised to a common format.
    """
    provider = get_api_provider(model.api)
    if not provider:
        msg = AssistantMessage(
            role="assistant", content=[], stop_reason="error", 
            provider=model.provider, model=model.id, api=model.api,
            error_message=f"No provider implementation for api={model.api!r}"
        )
        yield ErrorEvent(error=msg, reason="error")
        return
        
    async for event in _with_error_handling(provider.stream(model, context, options), model):
        yield event


async def complete(
    model: ModelDef,
    context: Context,
    options: Optional[StreamOptions] = None,
) -> AssistantMessage:
    """
    Non-streaming helper. Returns the final AssistantMessage.
    Useful for simple one-shot calls.
    """
    from .types import AssistantMessage, TextContent, ThinkingContent, ToolUseContent, TokenUsage, StopEvent, ErrorEvent

    blocks = []
    usage = TokenUsage()
    stop_reason = "stop"
    error_message = None

    async for event in stream(model, context, options):
        if event.type == "text_end":
            blocks.append(TextContent(text=event.text))
        elif event.type == "thinking_end":
            blocks.append(ThinkingContent(thinking=event.thinking))
        elif event.type == "tool_use_end":
            blocks.append(event.tool_use)
        elif event.type == "stop":
            usage = event.usage
            stop_reason = event.stop_reason
        elif event.type == "error":
            usage = event.error.usage
            blocks = event.error.content
            stop_reason = event.reason
            error_message = event.error.error_message
            break

    return AssistantMessage(
        role="assistant",
        content=blocks,
        stop_reason=stop_reason,
        usage=usage,
        provider=model.provider,
        model=model.id,
        api=model.api,
        error_message=error_message
    )


async def stream_simple(
    model: ModelDef,
    context: Context,
    options: Optional[StreamOptions] = None,
) -> AsyncIterator[StreamEvent]:
    """
    Unified Interface for reasoning models and standard streams.
    """
    options = options or StreamOptions()

    
    provider = get_api_provider(model.api)
    if not provider:
        yield ErrorEvent(error=AssistantMessage(
            role="assistant", content=[], stop_reason="error", provider=model.provider, model=model.id, api=model.api, error_message=f"No provider implementation for api={model.api!r}"
        ), reason="error")
        return
        
    # Allow providers to have explicit simple streaming handlers if configured, 
    # otherwise fallback to generic stream wrapper with injected options
    if hasattr(provider, "stream_simple") and provider.stream_simple != provider.stream:
        async for event in _with_error_handling(provider.stream_simple(model, context, options), model):
            yield event
    else:
        async for event in _with_error_handling(provider.stream(model, context, options), model):
            yield event


async def complete_simple(
    model: ModelDef,
    context: Context,
    options: Optional[StreamOptions] = None,
) -> AssistantMessage:
    """
    Non-streaming helper for unified reasoning. Returns final AssistantMessage.
    """
    from .types import AssistantMessage, TextContent, ThinkingContent, ToolUseContent, TokenUsage, StopEvent, ErrorEvent

    blocks = []
    usage = TokenUsage()
    stop_reason = "stop"
    error_message = None

    async for event in stream_simple(model, context, options):
        if event.type == "text_end":
            blocks.append(TextContent(text=event.text))
        elif event.type == "thinking_end":
            blocks.append(ThinkingContent(thinking=event.thinking))
        elif event.type == "tool_use_end":
            blocks.append(event.tool_use)
        elif event.type == "stop":
            usage = event.usage
            stop_reason = event.stop_reason
        elif event.type == "error":
            usage = event.error.usage
            blocks = event.error.content
            stop_reason = event.reason
            error_message = event.error.error_message
            break

    return AssistantMessage(
        role="assistant",
        content=blocks,
        stop_reason=stop_reason,
        usage=usage,
        provider=model.provider,
        model=model.id,
        api=model.api,
        error_message=error_message
    )



"""
Anthropic provider.
"""
from __future__ import annotations
import json
from typing import AsyncIterator, TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ..types import Context, ModelDef, StreamOptions

from ..types import (
    ToolUseContent, TokenUsage, ReasoningEffort,
    StartEvent, TextStartEvent, TextDeltaEvent, TextEndEvent,
    ThinkingStartEvent, ThinkingDeltaEvent, ThinkingEndEvent,
    ToolUseStartEvent, ToolUseDeltaEvent, ToolUseEndEvent,
    StopEvent, ErrorEvent, StreamEvent,
)
from ..transform_messages import to_anthropic


def _map_reasoning(effort: ReasoningEffort) -> dict | None:
    """Map ReasoningEffort → Anthropic thinking param."""
    if effort == ReasoningEffort.NONE:
        return None
    # xhigh clamped to high for Anthropic
    budget = {
        ReasoningEffort.LOW:    1_000,
        ReasoningEffort.MEDIUM: 5_000,
        ReasoningEffort.HIGH:   10_000,
        ReasoningEffort.XHIGH:  10_000,
    }.get(effort, 5_000)
    return {"type": "enabled", "budget_tokens": budget}



class AnthropicProvider:
    def __init__(self):
        self.api = "anthropic"

    async def stream(
        self,
        model: "ModelDef",
        context: "Context",
        options: Optional["StreamOptions"] = None,
    ) -> AsyncIterator[StreamEvent]:
        
        from anthropic import AsyncAnthropic
        from ..env import get_env_api_key
        
        api_key = get_env_api_key("anthropic")
        if not api_key:
             raise RuntimeError("No API key found for anthropic")

        # ... imports ...
        from ..transform_messages import to_anthropic, transform_messages
        
        # ... logic ...

        client_instance = AsyncAnthropic(api_key=api_key, base_url=model.base_url)
        
        reasoning = options.reasoning_effort if options else ReasoningEffort.NONE

        # 1. Transform messages
        transformed_messages = transform_messages(context.messages, model)

        params: dict = {
            "model":      model.id,
            "max_tokens": model.max_output,
            "messages":   to_anthropic(transformed_messages),
        }
        if context.system_prompt:
            params["system"] = context.system_prompt
        if context.tools:
            params["tools"] = [t.to_anthropic() for t in context.tools]
        thinking = _map_reasoning(reasoning)
        if thinking:
            params["thinking"] = thinking

        yield StartEvent(model=model.id)
        building: dict[str, dict] = {}
        text_buf  = ""
        think_buf = ""
        in_text   = False
        in_think  = False

        try:
            async with client_instance.messages.stream(**params) as s:
                async for ev in s:
                    t = ev.type

                    if t == "content_block_start":
                        cb = ev.content_block
                        if cb.type == "text":
                            in_text = True
                            yield TextStartEvent()
                        elif cb.type == "thinking":
                            in_think = True
                            yield ThinkingStartEvent()
                        elif cb.type == "tool_use":
                            building[cb.id] = {"id": cb.id, "name": cb.name, "raw": ""}
                            yield ToolUseStartEvent(id=cb.id, name=cb.name)

                    elif t == "content_block_delta":
                        d = ev.delta
                        if d.type == "text_delta":
                            text_buf += d.text
                            yield TextDeltaEvent(text=d.text)
                        elif d.type == "thinking_delta":
                            think_buf += d.thinking
                            yield ThinkingDeltaEvent(thinking=d.thinking)
                        elif d.type == "input_json_delta":
                            for tc in building.values():
                                tc["raw"] += d.partial_json
                                yield ToolUseDeltaEvent(id=tc["id"], partial_input=d.partial_json)

                    elif t == "content_block_stop":
                        if in_text:
                            yield TextEndEvent(text=text_buf)
                            text_buf = ""
                            in_text  = False
                        elif in_think:
                            yield ThinkingEndEvent(thinking=think_buf)
                            think_buf = ""
                            in_think  = False

                    elif t == "message_stop":
                        for tc in building.values():
                            try:
                                args = json.loads(tc["raw"]) if tc["raw"] else {}
                            except Exception:
                                args = {}
                            yield ToolUseEndEvent(
                                tool_use=ToolUseContent(id=tc["id"], name=tc["name"], arguments=args)
                            )

                        fm = await s.get_final_message()
                        u  = fm.usage
                        yield StopEvent(
                            stop_reason=fm.stop_reason or "end_turn",
                            usage=TokenUsage(
                                input_tokens       = u.input_tokens,
                                output_tokens      = u.output_tokens,
                                cache_read_tokens  = getattr(u, "cache_read_input_tokens", 0),
                                cache_write_tokens = getattr(u, "cache_creation_input_tokens", 0),
                            ),
                        )

        except Exception as e:
            raise RuntimeError(str(e))



import json
from json_repair import repair_json

def parse_streaming_json(partial_json: str | None) -> dict:
    """
    Attempts to parse potentially incomplete JSON during streaming.
    Always returns a valid object, even if the JSON is incomplete.
    """
    if not partial_json or not partial_json.strip():
        return {}
    
    # Try standard parsing first
    try:
        return json.loads(partial_json)
    except json.JSONDecodeError:
        # Try repairing
        try:
            repaired = repair_json(partial_json, return_objects=True)
            if isinstance(repaired, dict):
                return repaired
            raise ValueError(
                f"Repaired JSON is not an object. Received type: {type(repaired)}. "
                f"Partial: {partial_json[:200]}..."
            )
        except Exception as e:
            raise ValueError(
                f"Failed to parse JSON even with repair: {e}\n"
                f"Partial JSON: {partial_json[:200]}..."
            )



import pytest

from sentarc_ai.registry import ApiProvider, get_api_provider, register_api_provider
import sentarc_ai.registry as registry_module


class DummyProvider:
    async def stream(self, model, context, options=None):
        if False:
            yield None


def test_get_api_provider_resolves_aliases():
    original_registry = dict(registry_module._registry)
    try:
        registry_module._registry.clear()
        register_api_provider(ApiProvider(api="openai", stream=DummyProvider().stream))
        register_api_provider(ApiProvider(api="google", stream=DummyProvider().stream))
        register_api_provider(ApiProvider(api="anthropic", stream=DummyProvider().stream))

        assert get_api_provider("openai") is not None
        assert get_api_provider("openai-completions") is not None
        assert get_api_provider("google-generative-ai") is not None
        assert get_api_provider("anthropic-messages") is not None
        assert get_api_provider("unknown-api") is None
    finally:
        registry_module._registry.clear()
        registry_module._registry.update(original_registry)


def test_register_api_provider_validates_stream_callable():
    original_registry = dict(registry_module._registry)
    try:
        registry_module._registry.clear()
        bad = ApiProvider(api="openai", stream=DummyProvider().stream)
        bad.stream = None
        with pytest.raises(TypeError, match="stream must be callable"):
            register_api_provider(bad)
    finally:
        registry_module._registry.clear()
        registry_module._registry.update(original_registry)


def test_register_api_provider_validates_api_name():
    original_registry = dict(registry_module._registry)
    try:
        registry_module._registry.clear()
        bad = ApiProvider(api="openai", stream=DummyProvider().stream)
        bad.api = ""
        with pytest.raises(ValueError, match="must be a non-empty string"):
            register_api_provider(bad)
    finally:
        registry_module._registry.clear()
        registry_module._registry.update(original_registry)



import pytest
from sentarc_ai.utils.sanitize_unicode import sanitize_surrogates

def test_sanitize_surrogates():
    # Valid text should remain unchanged
    assert sanitize_surrogates("Hello World") == "Hello World"
    
    # Valid emojis (using proper Unicode encoding, not loose surrogates strings)
    assert sanitize_surrogates("Hello 🙈 World") == "Hello 🙈 World"
    
    # Isolated high surrogate (e.g., D83D from an improperly split string)
    assert sanitize_surrogates("Hello \ud83d World") == "Hello  World"
    assert sanitize_surrogates("Start \ud83d") == "Start "
    
    # Isolated low surrogate (e.g., DE48)
    assert sanitize_surrogates("Hello \ude48 World") == "Hello  World"
    assert sanitize_surrogates("\ude48 End") == " End"
    
    # Multiple isolated surrogates sprinkled in text
    assert sanitize_surrogates("\ud83d Isolated \ude48 surrogates \ud83d everywhere \ude48") == " Isolated  surrogates  everywhere "
    
    # Ensure a valid surrogate pair written with escapes survives if Python parses it as a full char
    text = "Valid \ud83d\ude48 Pair"
    assert sanitize_surrogates(text) == text
    
    # Edge cases
    assert sanitize_surrogates("") == ""
    assert sanitize_surrogates("\ud800") == ""  # Lowest high surrogate
    assert sanitize_surrogates("\udbff") == ""  # Highest high surrogate
    assert sanitize_surrogates("\udc00") == ""  # Lowest low surrogate
    assert sanitize_surrogates("\udfff") == ""  # Highest low surrogate



import pytest

from sentarc_ai.providers.openai_utils import convert_responses_messages, process_responses_stream
from sentarc_ai.types import (
    Context,
    ImageContent,
    Message,
    ModelDef,
    Role,
    TextContent,
    ToolResultMessage,
)


def test_convert_responses_messages_tool_role_and_vision_flag():
    model = ModelDef(id="o4-mini", provider="openai", api="openai-responses", supports_vision=False)
    context = Context(
        system_prompt="You are helpful",
        messages=[
            Message(role=Role.USER, content=[TextContent(text="hello")]),
            ToolResultMessage(
                role=Role.TOOL,
                tool_call_id="call_123|fc_123",
                content=[
                    TextContent(text="done"),
                    ImageContent(media_type="image/png", data="abc"),
                ],
            ),
        ],
    )

    out = convert_responses_messages(model, context, allowed_tool_call_providers={"openai"})

    assert out[0]["role"] == "system"
    tool_outputs = [m for m in out if isinstance(m, dict) and m.get("type") == "function_call_output"]
    assert len(tool_outputs) == 1
    assert tool_outputs[0]["call_id"] == "call_123"
    assert tool_outputs[0]["output"] == "done"
    # supports_vision=False should avoid appending extra input_image user content
    assert not any(
        isinstance(m, dict)
        and m.get("role") == "user"
        and any(isinstance(p, dict) and p.get("type") == "input_image" for p in m.get("content", []))
        for m in out
    )


@pytest.mark.asyncio
async def test_process_responses_stream_completed_usage_without_total_tokens_field():
    async def fake_stream():
        yield {
            "type": "response.completed",
            "response": {
                "status": "completed",
                "usage": {
                    "input_tokens": 12,
                    "output_tokens": 7,
                    "input_tokens_details": {"cached_tokens": 3},
                },
            },
        }

    model = ModelDef(id="o4-mini", provider="openai", api="openai-responses")
    events = [e async for e in process_responses_stream(fake_stream(), model)]

    assert events[0].type == "start"
    assert events[-1].type == "stop"
    assert events[-1].usage.input_tokens == 12
    assert events[-1].usage.output_tokens == 7
    assert events[-1].usage.cache_read_tokens == 3
